# Copyright 2019 Sean Robertson

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''Utility functions'''

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import warnings

import torch
import pydrobert.torch

__author__ = "Sean Robertson"
__email__ = "sdrobert@cs.toronto.edu"
__license__ = "Apache 2.0"
__copyright__ = "Copyright 2019 Sean Robertson"
__all__ = [
    'beam_search_advance',
    'error_rate',
    'optimal_completion',
    'optimizer_to',
    'prefix_error_rates',
    'random_walk_advance',
]


def beam_search_advance(
        logits_t, width, log_prior=None, y_prev=None,
        eos=pydrobert.torch.INDEX_PAD_VALUE, lens=None, prevent_eos=False):
    r'''Advance a beam search

    Suppose a model outputs a un-normalized log-probability distribution over
    the next element of a sequence in `logits_t` s.t.

    .. math::

        Pr(y_t = c; log\_prior) = exp(logits_{t,c}) / \sum_k exp(logits_{t,k})

    We assume :math:`logits_t` is a function of what comes before
    :math:`logits_t = f(logits_{<t}, y_{<t})`. Alternatively, letting
    :math:`s_t = (logits_t, y_t)`, :math:`s` is a Markov Chain. A model is
    auto-regressive if :math:`f` depends on :math:`y_{<t}`, and is not
    auto-regressive if :math:`logits_t = f(logits_{<t})`.

    Beam search is a heuristic mechanism for determining a best path, i.e.
    :math:`\arg \max_y Pr(y)` that maximizes the probability of the best path
    by keeping track of `width` high probability paths called "beams" (the
    aggregate of which for a given batch element is named, unfortunately, "the
    beam"). If the model is auto-regressive, beam search is only approximate.
    However, if the model is not auto-regressive, beam search gives an exact
    n-best list.

    This function is called at every time step. It updates old beam
    log-probabilities (`log_prior`) with new ones (`score`) from the joint
    distribution with `logits`, and updates us the class indices emitted
    between them (`y`). See the examples section for how this might work.

    Parameters
    ----------
    logits : torch.FloatTensor
        The conditional probabilities over class labels for the current time
        step. Either of shape ``(num_batches, old_width, num_classes)``,
        where ``old_width`` is the number of beams in the previous time step,
        or ``(num_batches, num_classes)``, where it is assumed that
        ``old_width == 1``
    width : int
        The number of beams in the beam to produce for the current time step.
        ``width <= num_classes``
    log_prior : torch.FloatTensor, optional
        A tensor of (or proportional to) log prior probabilities of beams up
        to the previous time step. Either of shape ``(num_batches, old_width)``
        or ``(num_batches,)``. If unspecified, a uniform log prior will be used
    y_prev : torch.LongTensor, optional
        A tensor of shape ``(t - 1, num_batches, old_width)`` or
        ``(t - 1, num_batches)`` specifying :math:`y_{<t}`. If unspecified,
        it is assumed that ``t == 1``
    eos : int, optional
        A special end-of-sequence symbol indicating that the beam has ended.
        Can be a class index. If this value occurs in in ``y_prev[-1, bt, bm]``
        for some batch ``bt`` and beam ``bm``, that beam will be padded with
        an `eos` token and the score for that beam won't change
    lens : torch.LongTensor, optional
        A tensor of shape ``(num_batches,)``. If ``t > lens[bt]`` for some
        batch ``bt``, all beams for ``bt`` will be considered finished. All
        scores will be fixed and `eos` will be appended to `y_prev`
    prevent_eos : bool, optional
        Setting this flag to :obj:`True` will keep `eos` targets from entering
        a beam unless it has finished (either with a prior `eos` or through
        `lens`). Note that this will only have an effect when ``0 <= eos <=
        num_classes``

    Returns
    -------
    score : torch.FloatTensor
        Of shape ``(num_batches, width)`` of the log-joint probabilities of the
        new beams in the beam
    y : torch.LongTensor
        Of shape ``(t, num_batches, width)`` of indices of the class labels
        generated up to this point
    s : torch.LongTensor
        Of shape ``(num_batches, width)`` of indices of beams in the old beam
        which prefix the new beam. Note that beams in the new beam are sorted
        by descending probability

    Examples
    --------

    Auto-regressive decoding with beam search. We assume that all input have
    the same number of steps

    >>> N, I, C, T, W, H, eos, start = 5, 5, 10, 100, 5, 10, 0, -1
    >>> cell = torch.nn.RNNCell(I + 1, H)
    >>> ff = torch.nn.Linear(H, C)
    >>> inp = torch.rand(T, N, I)
    >>> y = torch.full((1, N, 1), start, dtype=torch.long)
    >>> h_t = torch.zeros(N, 1, H)
    >>> score = None
    >>> for inp_t in inp:
    >>>     y_tm1 = y[-1]
    >>>     old_width = y_tm1.shape[-1]
    >>>     inp_t = inp_t.unsqueeze(1).expand(N, old_width, I)
    >>>     x_t = torch.cat([inp_t, y_tm1.unsqueeze(2).float()], -1)
    >>>     h_t = cell(
    ...         x_t.view(N * old_width, I + 1),
    ...         h_t.view(N * old_width, H)
    ...     ).view(N, old_width, H)
    >>>     logits_t = ff(h_t)
    >>>     score, y, s_t = beam_search_advance(logits_t, W, score, y, eos)
    >>>     h_t = h_t.gather(1, s_t.unsqueeze(-1).expand(N, W, H))
    >>> bests = []
    >>> for batch_idx in range(N):
    >>>     best_beam_path = y[1:, batch_idx, 0]
    >>>     not_special_mask = best_beam_path.ne(eos)
    >>>     best_beam_path = best_beam_path.masked_select(not_special_mask)
    >>>     bests.append(best_beam_path)

    ``W``-best list for non-auto-regressive model. We don't emit an `eos`,
    instead completing the sequence when we've hit the target length via `lens`

    >>> N, I, C, T, W, H = 5, 5, 10, 100, 5, 10
    >>> rnn = torch.nn.RNN(I, H)
    >>> ff = torch.nn.Linear(H, C)
    >>> inp = torch.rand(T, N, I)
    >>> lens = torch.randint(1, T + 1, (N,)).sort(descending=True)[0]
    >>> packed_inp = torch.nn.utils.rnn.pack_padded_sequence(inp, lens)
    >>> packed_h, _ = rnn(packed_inp)
    >>> packed_logits = ff(packed_h[0])
    >>> logits = torch.nn.utils.rnn.pad_packed_sequence(
    ...     torch.nn.utils.rnn.PackedSequence(
    ...         packed_logits, batch_sizes=packed_h[1]),
    ...     total_length=T,
    ... )[0]
    >>> y = score = None
    >>> for t, logits_t in enumerate(logits):
    >>>     if t:
    >>>         logits_t = logits_t.unsqueeze(1).expand(-1, W, -1)
    >>>     score, y, _ = beam_search_advance(logits_t, W, score, y, lens=lens)
    '''
    if logits_t.dim() == 2:
        logits_t = logits_t.unsqueeze(1)
    elif logits_t.dim() != 3:
        raise ValueError('logits_t must have dimension of either 2 or 3')
    logits_t = torch.nn.functional.log_softmax(logits_t, 2)
    neg_inf = torch.tensor(-float('inf'), device=logits_t.device)
    num_batches, old_width, num_classes = logits_t.shape
    if log_prior is None:
        log_prior = torch.full(
            (num_batches, old_width),
            -torch.log(torch.tensor(float(num_classes))),
            dtype=logits_t.dtype, device=logits_t.device,
        )
    elif tuple(log_prior.shape) == (num_batches,) and old_width == 1:
        log_prior = log_prior.unsqueeze(1)
    elif log_prior.shape != logits_t.shape[:-1]:
        raise ValueError(
            'If logits_t of shape {} then log_prior must have shape {}'.format(
                (num_batches, old_width, num_classes),
                (num_batches, old_width),
            ))
    if prevent_eos and 0 <= eos < num_classes:
        # we have to put this before the num_done check so that it'll be
        # overwritten for paths that have finished already
        logits_t[..., eos] = neg_inf
    eos_set = None
    if y_prev is not None:
        if y_prev.dim() == 2:
            y_prev = y_prev.unsqueeze(2)
        if y_prev.shape[1:] != log_prior.shape:
            raise ValueError(
                'If logits_t of shape {} then y_prev must have shape '
                '(*, {}, {})'.format(
                    (num_batches, old_width, num_classes),
                    num_batches, old_width,
                )
            )
        eos_mask = y_prev[-1].eq(eos)
        num_done = eos_mask.sum(1)
        if num_done.sum().item():
            if old_width < width and torch.any(num_done == old_width):
                raise ValueError(
                    'New beam width ({}) is wider than old beam width '
                    '({}), but all paths are already done in one or more '
                    'batch elements'.format(width, old_width))
            # we're going to treat class 0 as the sentinel for eos (even if
            # eos is a legit class label)
            done_classes = torch.full_like(logits_t[0, 0], neg_inf)
            done_classes[0] = 0.
            logits_t = torch.where(
                eos_mask.unsqueeze(2),
                done_classes,
                logits_t,
            )
            # If eos_mask looks like this (vertical batch, horizontal beam):
            #   1 0 0 1 0
            #   0 1 0 0 0
            #   0 0 0 0 0
            # then eos_set will be
            #    0 -1 -1  3 -1
            #   -1  1 -1 -1 -1
            #   -1 -1 -1 -1 -1
            # s might look like
            #    1 2 3 3
            #    2 2 4 1
            #    1 2 3 4
            # we'll compare a single value from a row of s to a matched row
            # of eos_set. Any match means the beam had finished already. The
            # mask on y will be
            #    0 0 1 1
            #    0 0 0 1
            #    0 0 0 0
            # pretty funky
            eos_set = torch.where(
                eos_mask,
                torch.arange(old_width, device=logits_t.device),
                torch.tensor(-1, device=logits_t.device).expand(old_width),
            )
        t = y_prev.shape[0] + 1
    else:
        t = 1
    len_mask = None
    if lens is not None:
        if lens.shape != logits_t.shape[:1]:
            raise ValueError('lens must be of shape ({},)'.format(num_batches))
        len_mask = lens.lt(t)
        if torch.any(len_mask):
            if old_width < width:
                raise ValueError(
                    'New beam width ({}) is wider than old beam width '
                    '({}), but all paths are already done in one or more '
                    'batch elements'.format(width, old_width))
        else:
            len_mask = None
    joint = log_prior.unsqueeze(2) + logits_t
    score, idxs = torch.topk(joint.view(num_batches, -1), width, dim=1)
    s = idxs // num_classes
    y = (idxs % num_classes).unsqueeze(0)
    if eos_set is not None:
        y_mask = (s.unsqueeze(2) == eos_set.unsqueeze(1)).any(2)
        y = y.masked_fill(y_mask, eos)
    if len_mask is not None:
        score = torch.where(
            len_mask.unsqueeze(1),
            log_prior[..., :width],
            score,
        )
        y = y.masked_fill(len_mask.unsqueeze(1).unsqueeze(0), eos)
        s = torch.where(
            len_mask.unsqueeze(1),
            torch.arange(width, device=logits_t.device),
            s,
        )
    if y_prev is not None:
        y_prev = y_prev.gather(
            2, s.unsqueeze(0).expand(t - 1, num_batches, width))
        y = torch.cat([y_prev, y], 0)
    return score, y, s


def error_rate(
        ref, hyp, eos=None, include_eos=False, norm=True, batch_first=False,
        ins_cost=1., del_cost=1., sub_cost=1., warn=True):
    '''Calculate error rates over a batch

    An error rate is merely a `Levenshtein (edit) distance
    <https://en.wikipedia.org/wiki/Levenshtein_distance>`__ normalized over
    reference sequence lengths.

    Given a reference (gold-standard) transcript tensor `ref` of size
    ``(max_ref_steps, batch_size)`` if ``batch_first == False`` or
    ``(batch_size, max_hyp_steps)`` otherwise, and a similarly shaped tensor of
    hypothesis transcripts `hyp`, this function produces a tensor `er` of shape
    ``(batch_size,)`` storing the associated error rates.

    `er` will not have a gradient, and is thus not directly suited to being a
    loss function

    Parameters
    ----------
    ref : torch.LongTensor
    hyp : torch.LongTensor
    eos : int, optional
        A special token in `ref` and `hyp` whose first occurrence in each
        batch indicates the end of a transcript. This allows for
        variable-length transcripts in the batch
    include_eos : bool, optional
        Whether to include the first instance of `eos` found in both `ref` and
        `hyp` as valid tokens to be computed as part of the distance. This is
        useful when gauging if a model is learning to emit the `eos` properly,
        but is not usually included in an evaluation. Only the first `eos` per
        transcript is included
    norm : bool, optional
        If :obj:`False`, will return edit distances instead of error rates
    batch_first : bool, optional
    ins_cost : float, optional
        The cost of an adding a superfluous token to a transcript in `hyp`
    del_cost : float, optional
        The cost of missing a token from `ref`
    sub_cost : float, optional
        The cost of swapping a token from `ref` with one from `hyp`
    warn : bool, optional
        Whether to display warnings on irregularities. Currently, this can
        happen in two ways:

        1. If :obj:`True` and `norm` is :obj:`True`, will warn when a reference
           transcription has zero length
        2. If `eos` is set and `include_eos` is :obj:`True`, will warn when a
           transcript does not include an `eos` symbol

    Returns
    -------
    er : torch.FloatTensor
    '''
    er = _levenshtein(
        ref, hyp, eos, include_eos, batch_first, ins_cost,
        del_cost, sub_cost, warn, norm=norm)
    return er


def optimal_completion(
        ref, hyp, eos=None, include_eos=True, batch_first=False, ins_cost=1.,
        del_cost=1., sub_cost=1., padding=pydrobert.torch.INDEX_PAD_VALUE,
        exclude_last=False, warn=True):
    r'''Return a mask of next tokens of a minimum edit distance prefix

    Given a reference transcript `ref` of shape ``(max_ref_steps, batch_size)``
    (or ``(batch_size, max_ref_steps)`` if `batch_first` is :obj:`True`) and a
    hypothesis transcript `hyp` of shape ``(max_hyp_steps, batch_size)``, (or
    ``(batch_size, max_hyp_steps)``), this function produces a long tensor
    `optimals` of shape ``(max_hyp_steps + 1, batch_size, max_unique_next)``
    (or ``(batch_size, max_hyp_steps + 1, max_unique_next)``), where
    ``max_unique_next <= max_ref_steps``, of the unique tokens that could be
    added to the hypothesis prefix ``hyp[:prefix_len, batch]`` such that some
    remaining suffix concatenated to the prefix would result in a minimal edit
    distance. See below for an example

    Parameters
    ----------
    ref : torch.LongTensor
    hyp : torch.LongTensor
    eos : int, optional
        A special token in `ref` and `hyp` whose first occurrence in each
        batch indicates the end of a transcript. This allows for
        variable-length transcripts in the batch
    include_eos : bool, optional
        Whether to include the first instance of `eos` found in both `ref` and
        `hyp` as valid tokens to be computed as part of the distance and next
        tokens for a suffix. Only the first `eos` per transcript is included
    batch_first : bool, optional
    ins_cost : float, optional
        The cost of an adding a superfluous token to a transcript in `hyp`
    del_cost : float, optional
        The cost of missing a token from `ref`
    sub_cost : float, optional
        The cost of swapping a token from `ref` with one from `hyp`
    padding : int, optional
        The value to right-pad unequal-length sequences with
    exclude_last : bool, optional
        If true, will exclude the final prefix, consisting of the entire
        transcript, from the returned `optimals`. Optimals will be of shape
        ``(max_hyp_steps, batch_size, max_unique_next)``
    warn : bool, optional
        Whether to display warnings on irregularities. Currently, this only
        occurs when `eos` is set, `include_eos` is :obj:`True`, and a
        transcript does not contain the `eos` symbol

    Returns
    -------
    optimals : torch.LongTensor

    Examples
    --------

    Consider the reference text "foot" and the hypothesis text "bot". The below
    shows the matrix used to calculate edit distances between them::

        \ _ f o o t
        _ 0 1 2 3 4
        b 1 1 2 3 4
        o 2 2 1 2 3
        t 3 3 2 2 2

    If ``prefix_len == 0``, then the prefix is "", and "f" (from the suffix
    "foot") is the only subsequent token that would not increase the edit
    distance from that of the prefix (0). If ``prefix_len == 1``, then the
    prefix is "b". To arrive at the minimum edit distance for "b", one either
    treats "b" as an insertion or a substitution for "f", yielding suffixes
    "foot" and "oot". Thus, the subsequent token could be "f" or "o". For the
    prefix "bo", the minimum edit distance is achieved by first substituting
    "f" for "b", then substituting "o" for "o", resulting in the suffix "ot"
    and the next optimal character "o". Finally, for ``prefix_len == 3`` and
    prefix "bot", there are many operations that can produce the minimum edit
    distance of 2, resulting in one of the suffixes "ot", "t", and "". The
    latter suffix requires no more tokens and so any operation would increase
    the edit distance. Thus the optimal next tokens could be "o" or "t".

    Plugging "foot" and "bot" into this function, we get the prefixes:

    >>> ref_text, hyp_text = "foot", "bot"
    >>> ref = torch.tensor([ord(c) for c in ref_text]).unsqueeze(1)
    >>> hyp = torch.tensor([ord(c) for c in hyp_text]).unsqueeze(1)
    >>> optimal = optimal_completion(ref, hyp).squeeze(1)
    >>> for prefix_len, o_for_pr in enumerate(optimal):
    ...     o_for_pr = o_for_pr.masked_select(o_for_pr.ge(0)).tolist()
    ...     print('prefix={}: {}'.format(
    ...         hyp_text[:prefix_len], ','.join([chr(i) for i in o_for_pr])))
    prefix=: f
    prefix=b: f,o
    prefix=bo: o
    prefix=bot: o,t

    See Also
    --------
    pydrobert.torch.training.HardOptimalCompletionDistillationLoss
        A loss function that uses these optimal completions to train a model
    '''
    mask = _levenshtein(
        ref, hyp, eos, include_eos, batch_first, ins_cost, del_cost,
        sub_cost, warn, return_mask=True, exclude_last=exclude_last,
    )
    max_hyp_steps_p1, max_ref_steps, batch_size = mask.shape
    targets = []
    if batch_first:
        for mask_bt, ref_bt in zip(mask.transpose(0, 2), ref):
            for mask_bt_hyp in mask_bt.t():
                targets.append(torch.unique(ref_bt.masked_select(mask_bt_hyp)))
    else:
        for mask_hyp in mask:
            for mask_hyp_bt, ref_bt in zip(mask_hyp.t(), ref.t()):
                targets.append(torch.unique(ref_bt.masked_select(mask_hyp_bt)))
    targets = torch.nn.utils.rnn.pad_sequence(
        targets, padding_value=padding, batch_first=True)
    if batch_first:
        targets = targets.view(batch_size, max_hyp_steps_p1, -1)
    else:
        targets = targets.view(max_hyp_steps_p1, batch_size, -1)
    return targets


def optimizer_to(optimizer, to):
    '''Move tensors in an optimizer to another device

    This function traverses the state dictionary of an `optimizer` and pushes
    any tensors it finds to the device implied by `to` (as per the semantics
    of :func:`torch.Tensor.to`)
    '''
    state_dict = optimizer.state_dict()
    key_stack = [(x,) for x in state_dict.keys()]
    new_device = False
    while key_stack:
        last_dict = None
        val = state_dict
        keys = key_stack.pop()
        for key in keys:
            last_dict = val
            val = val[key]
        if isinstance(val, torch.Tensor):
            try:
                last_dict[keys[-1]] = val.to(to)
                if last_dict[keys[-1]].device != val.device:
                    new_device = True
            except Exception as e:
                print(e)
                pass
        elif hasattr(val, 'keys'):
            key_stack += [keys + (x,) for x in val.keys()]
    if new_device:
        optimizer.load_state_dict(state_dict)


def prefix_error_rates(
        ref, hyp, eos=None, include_eos=True, norm=True, batch_first=False,
        ins_cost=1., del_cost=1., sub_cost=1.,
        padding=pydrobert.torch.INDEX_PAD_VALUE, exclude_last=False,
        warn=True):
    '''Compute the error rate between ref and each prefix of hyp

    Given a reference transcript `ref` of shape ``(max_ref_steps, batch_size)``
    (or ``(batch_size, max_ref_steps)`` if `batch_first` is :obj:`True`) and a
    hypothesis transcript `hyp` of shape ``(max_hyp_steps, batch_size)``, (or
    ``(batch_size, max_hyp_steps)``), this function produces a tensor
    `prefix_ers` of shape ``(max_hyp_steps + 1, batch_size)`` (or
    ``(batch_size, max_hyp_steps + 1))`` which contains the error rates for
    each prefix of each hypothesis, starting from the empty prefix

    Parameters
    ----------
    ref : torch.LongTensor
    hyp : torch.LongTensor
    eos : int, optional
        A special token in `ref` and `hyp` whose first occurrence in each
        batch indicates the end of a transcript. This allows for
        variable-length transcripts in the batch
    include_eos : bool, optional
        Whether to include the first instance of `eos` found in both `ref` and
        `hyp` as valid tokens to be computed as part of the distance.
        Only the first `eos` per transcript is included
    norm : bool, optional
        If :obj:`False`, will return edit distances instead of error rates
    batch_first : bool, optional
    ins_cost : float, optional
        The cost of an adding a superfluous token to a transcript in `hyp`
    del_cost : float, optional
        The cost of missing a token from `ref`
    sub_cost : float, optional
        The cost of swapping a token from `ref` with one from `hyp`
    padding : int, optional
        The value to right-pad the error rates of unequal-length sequences with
        in `prefix_ers`
    exclude_last : bool, optional
        If true, will exclude the final prefix, consisting of the entire
        transcript, from the returned `dists`. `dists` will be of shape
        ``(max_hyp_steps, batch_size, max_unique_next)``
    warn : bool, optional
        Whether to display warnings on irregularities. Currently, this only
        occurs when `eos` is set, `include_eos` is :obj:`True`, and a
        transcript does not contain the `eos` symbol

    Returns
    -------
    prefix_ers : torch.tensor

    See Also
    --------
    :ref:`Gradient Estimators`
        Provides an example where this function is used to determine a reward
        function for reinforcement learning
    '''
    prefix_ers = _levenshtein(
        ref, hyp, eos, include_eos, batch_first, ins_cost,
        del_cost, sub_cost, warn, norm=norm, return_prf_dsts=True,
        exclude_last=exclude_last, padding=padding,
    )
    return prefix_ers


def random_walk_advance(
        logits_t, num_samp, y_prev=None, eos=pydrobert.torch.INDEX_PAD_VALUE,
        lens=None, prevent_eos=False, include_relaxation=False):
    r'''Advance a random walk of sequences

    Suppose a model outputs a un-normalized log-probability distribution over
    the next element of a sequence in `logits_t` s.t.

    .. math::

        Pr(y_t = c) = exp(logits_{t,c}) / \sum_k exp(logits_k)

    We assume :math:`logits_t` is a function of what comes before
    :math:`logits_t = f(logits_{<t}, y_{<t})`. Alternatively, letting
    :math:`s_t = (logits_t, y_t)`, :math:`s` is a Markov Chain. A model is
    auto-regressive if :math:`f` depends on :math:`y_{<t}`, and is not
    auto-regressive if :math:`logits_t = f(logits_{<t})`.

    A random walk can be performed over a Markov Chain by sampling the elements
    :math:`y_t` of the greater sequence `y` one at a time, according to
    :math:`Pr(y_t = c)`. This allows us to sample the distribution
    :math:`Pr(Y)`.

    This function is called at every time step. It updates the sequences
    being built (`y_prev`) with one additional token and returns `y`. This
    function is intended to be coupled with an auto-regressive model, where
    `logits_t` is not known until :math:`y_t` is known. If the model is
    not auto-regressive, it is much more efficient to gather all `logits_t`
    into one :math:`logits` and sample all at once. See the examples section
    below for both behaviours

    Parameters
    ----------
    logits_t : torch.FloatTensor
        The conditional probabilities over class labels for the current time
        step. Either of shape ``(num_batches, old_samp, num_classes)``,
        where ``old_samp`` is the number of samples in the previous time
        step, or ``(num_batches, num_classes)``, where it is assumed that
        ``old_samp == 1``
    num_samp : int
        The number of samples to be drawn. Either ``old_samp == 1`` and/or
        ``num_samp <= old_samp`` must be :obj:`True`. That is, either all
        samples will share the same prefix, or we are building off a subset of
        the samples from ``y_prev`` (in this case, always the first `num_samp`)
    y_prev : torch.LongTensor, optional
        A tensor of shape ``(t - 1, num_batches, old_samp)`` or
        ``(t - 1, num_batches)`` specifying :math:`y_{<t}`. If unspecified,
        it is assumed that ``t == 1``
    eos : int, optional
        A special end-of-sequence symbol indicating that the beam has ended.
        Can be a class index. If this value occurs in in
        ``y_prev[-1, bt, smp]`` for some batch ``bt`` and sample ``smp``,
        `eos` will be appended to ``y_prev[:, bt, smp]``
    lens : torch.LongTensor, optional
        A tensor of shape ``(num_batches,)``. If ``t > lens[bt]`` for some
        batch ``bt``, all samples for ``bt`` will be considered finished. `eos`
        will be appended to `y_prev`
    prevent_eos : bool, optional
        Setting this flag to :obj:`True` will keep `eos` targets from being
        drawn unless a sample has finished (either with a prior `eos` or
        through `lens`). Note that this will only have an effect when ``0 <=
        eos <= num_classes``
    include_relaxation : bool, optional
        If :obj:`True`, a tuple will be returned whose second element is `z`,
        see below


    Returns
    -------
    y : torch.LongTensor
        A long tensor of shape ``(t, num_batches, num_samp)`` of the sampled
        sequences so far. Note that, since :math:`y_t` are drawn `i.i.d.`,
        there is no guarantee of the uniqueness of each `num_samp` samples
    z : torch.FloatTensor
        Only included if `include_relaxation` is :obj:`True`. `z` is a sample
        of a continuous relaxation of the categorical distribution of `logits`
        of shape ``(num_batches, num_samp, num_classes). Assuming ``y_prev[-1,
        bt, smp] != eos``, ``y[-1, bt, smp] == z[bt, smp].argmax(dim-1)``. If
        ``y_prev[-1, bt, smp] == eos``, ``z[bt, smp, :] = -infinity``. The
        primary purpose of `z` is to be used as an argument (alongside `y`) in
        more complicated gradient estimators from
        :mod:`pydrobert.torch.estimators`

    Examples
    --------

    Here is an example of random path sampling with a non-auto-regressive
    RNN. It does not need this function, and can take advantage of packed
    sequences for efficiency and gradient validity.

    >>> N, I, C, T, W, H, eos = 5, 4, 10, 100, 6, 15, 0
    >>> rnn = torch.nn.RNN(I, H)
    >>> ff = torch.nn.Linear(H, C)
    >>> inp = torch.rand(T, N, I)
    >>> lens = torch.randint(1, T + 1, (N,)).sort(descending=True)[0]
    >>> packed_inp = torch.nn.utils.rnn.pack_padded_sequence(inp, lens)
    >>> packed_h, _ = rnn(packed_inp)
    >>> packed_logits = ff(packed_h[0])
    >>> packed_logits_dup = packed_logits.detach().unsqueeze(1)
    >>> packed_logits_dup = packed_logits_dup.expand(-1, W, -1)  # (flat, W, C)
    >>> packed_y = torch.distributions.Categorical(
    ...     logits=packed_logits_dup).sample()  # (flat, W)
    >>> # we pad y with "eos" to ensure each sample is done by its length,
    >>> # but "eos" may have occurred beforehand
    >>> y = torch.nn.utils.rnn.pad_packed_sequence(
    ...     torch.nn.utils.rnn.PackedSequence(
    ...         packed_y, batch_sizes=packed_h[1]),
    ...     padding_value=eos, total_length=T,
    ... )[0]  # (T, N, W) (batch index gets inserted as 2nd dim)

    Here is an auto-regressive RNN that uses this function to build partial
    samples into `y`

    >>> N, I, C, T, W, H, eos, start = 5, 5, 10, 100, 5, 10, 0, -1
    >>> cell = torch.nn.RNNCell(I + 1, H)
    >>> ff = torch.nn.Linear(H, C)
    >>> inp = torch.rand(T, N, I)
    >>> y = torch.full((1, N, 1), start, dtype=torch.long)
    >>> h_t = torch.zeros(N, 1, H)
    >>> for inp_t in inp:
    >>>     y_tm1 = y[-1]
    >>>     old_samp = y_tm1.shape[-1]
    >>>     inp_t = inp_t.unsqueeze(1).expand(N, old_samp, I)
    >>>     x_t = torch.cat([inp_t, y_tm1.unsqueeze(2).float()], -1)
    >>>     h_t = cell(
    ...         x_t.view(N * old_samp, I + 1),
    ...         h_t.view(N * old_samp, H),
    ...     ).view(N, old_samp, H)
    >>>     logits_t = ff(h_t)
    >>>     y = random_walk_advance(logits_t, W, y, eos)
    >>>     if old_samp == 1:
    >>>         h_t = h_t.expand(-1, W, H).contiguous()

    See Also
    --------
    :ref:`Gradient Estimators`
        Includes a use case for `include_relaxation`
    '''
    if logits_t.dim() == 2:
        logits_t = logits_t.unsqueeze(1)
    elif logits_t.dim() != 3:
        raise ValueError('logits_t must have dimension of either 2 or 3')
    num_batches, old_samp, num_classes = logits_t.shape
    if prevent_eos and 0 <= eos < num_classes:
        logits_t[..., eos] = torch.tensor(
            -float('inf'), device=logits_t.device)
    if old_samp != 1 and num_samp > old_samp:
        raise ValueError(
            'either old_samp == 1 or num_samp <= old_samp must be true')
    eos_mask = None
    if y_prev is not None:
        if y_prev.dim() == 2:
            y_prev = y_prev.unsqueeze(2)
        if y_prev.shape[1:] != logits_t.shape[:-1]:
            raise ValueError(
                'If logits_t of shape {} then y_prev must have shape '
                '(*, {}, {})'.format(
                    (num_batches, old_samp, num_classes),
                    num_batches, old_samp,
                )
            )
        y_prev = y_prev.expand(-1, -1, num_samp)
        eos_mask = y_prev[-1].eq(eos)
        if eos_mask.any():
            eos_mask = eos_mask[..., :num_samp]
        else:
            eos_mask = None
        t = y_prev.shape[0] + 1
    else:
        t = 1
    logits_t = logits_t.expand(-1, num_samp, -1)
    if lens is not None:
        if lens.shape != logits_t.shape[:1]:
            raise ValueError('lens must be of shape ({},)'.format(num_batches))
        len_mask = lens.lt(t)
        if torch.any(len_mask):
            len_mask = len_mask.unsqueeze(1).expand(-1, num_samp)
            eos_mask = len_mask if eos_mask is None else (eos_mask | len_mask)
    u = torch.distributions.utils.clamp_probs(torch.rand_like(logits_t))
    log_theta = torch.nn.functional.log_softmax(logits_t, dim=-1)
    z = log_theta - torch.log(-torch.log(u))
    y = z.argmax(dim=-1)
    if eos_mask is not None:
        y = y.masked_fill(eos_mask, eos)
        z = z.masked_fill(eos_mask.unsqueeze(-1), -float('inf'))
    y = y.unsqueeze(0)
    if y_prev is not None:
        y = torch.cat([y_prev, y], 0)
    if include_relaxation:
        return y, z
    else:
        return y


def _levenshtein(
        ref, hyp, eos, include_eos, batch_first, ins_cost, del_cost,
        sub_cost, warn, norm=False, return_mask=False,
        return_prf_dsts=False, exclude_last=False, padding=None):
    assert not return_mask or not return_prf_dsts
    if ref.dim() != 2 or hyp.dim() != 2:
        raise ValueError('ref and hyp must be 2 dimensional')
    if batch_first:
        ref = ref.t()
        hyp = hyp.t()
    ref = ref.detach()
    hyp = hyp.detach()
    max_ref_steps, batch_size = ref.shape
    max_hyp_steps, batch_size_ = hyp.shape
    device = ref.device
    if batch_size != batch_size_:
        raise ValueError(
            'ref has batch size {}, but hyp has {}'.format(
                batch_size, batch_size_))
    if eos is not None:
        ref_lens = torch.full_like(ref[0], max_ref_steps)
        hyp_lens = torch.full_like(hyp[0], max_hyp_steps)
        for coord in ref.eq(eos).nonzero():
            ref_lens[..., coord[1]] = torch.min(ref_lens[coord[1]], coord[0])
        for coord in hyp.eq(eos).nonzero():
            hyp_lens[..., coord[1]] = torch.min(hyp_lens[coord[1]], coord[0])
        if include_eos:
            ref_eq_mask = ref_lens == max_ref_steps
            ref_lens = ref_lens + 1
            if ref_eq_mask.any():
                if warn:
                    warnings.warn(
                        "include_eos=True, but a transcription in ref did not "
                        "contain the eos symbol ({}). To suppress this "
                        "warning, set warn=False".format(eos))
                ref_lens = torch.where(
                    ref_eq_mask,
                    ref_lens - 1,
                    ref_lens
                )
            hyp_eq_mask = hyp_lens == max_hyp_steps
            hyp_lens = hyp_lens + 1
            if hyp_eq_mask.any():
                if warn:
                    warnings.warn(
                        "include_eos=True, but a transcription in hyp did not "
                        "contain the eos symbol ({}). To suppress this "
                        "warning, set warn=False".format(eos))
                hyp_lens = torch.where(
                    hyp_eq_mask,
                    hyp_lens - 1,
                    hyp_lens
                )
            del ref_eq_mask, hyp_eq_mask
    else:
        ref_lens = torch.full_like(ref[0], max_ref_steps)
        hyp_lens = torch.full_like(hyp[0], max_hyp_steps)
    ins_cost = torch.tensor(float(ins_cost), device=device)
    del_cost = torch.tensor(float(del_cost), device=device)
    sub_cost = torch.tensor(float(sub_cost), device=device)
    zero = torch.tensor(0., device=device)
    batch_range = torch.arange(batch_size, device=device)
    if return_mask:
        mask = torch.empty(
            (
                max_hyp_steps + (0 if exclude_last else 1),
                max_ref_steps, batch_size),
            device=device, dtype=torch.uint8)
        mask[0, 0] = 1
        mask[0, 1:] = 0
    elif return_prf_dsts:
        assert padding is not None
        prefix_ers = torch.empty(
            (max_hyp_steps + (0 if exclude_last else 1), batch_size),
            device=device, dtype=torch.float,
        )
        prefix_ers[0] = ref_lens
    # direct row down corresponds to insertion
    # direct col right corresponds to a deletion
    row = torch.arange(
        max_ref_steps + 1, device=device, dtype=torch.float
    ).unsqueeze(1).expand(max_ref_steps + 1, batch_size)
    last_row = torch.empty_like(row)
    # we vectorize as much as we can. Neither substitutions nor insertions
    # require values from the current row to be computed, and since the last
    # row can't be altered, we can easily vectorize there. We can't do the same
    # with deletions because they rely on what came before in the row
    for hyp_idx in range(1, max_hyp_steps + 1):
        last_row = row
        row = torch.where(
            hyp_lens < hyp_idx,
            last_row,
            last_row + ins_cost
        )
        sub_row = torch.where(
            ref == hyp[hyp_idx - 1],
            last_row[:-1],
            last_row[:-1] + sub_cost,
        )
        row[1:] = torch.min(row[1:], sub_row)
        for ref_idx in range(1, max_ref_steps + 1):
            row[ref_idx] = torch.min(row[ref_idx], row[ref_idx - 1] + del_cost)
        if return_mask and (hyp_idx < max_hyp_steps or not exclude_last):
            # As proven in the OCD paper, the optimal targets are always the
            # first character of a suffix of the reference transcript that
            # remains to be aligned. The levenshtein operation
            # corresponding to what we do with that target would be a matched
            # substitution (i.e. hyp's next token is the OCD target, resulting
            # in no change in cost from the prefix). Thus, given a levenshtein
            # matrix for one of these OCD targets (which is this matrix,
            # except for the final row), the minimal values on the final row
            # sit on a diagonal from the minimal values of the current row.
            mins = row.min(0, keepdim=True)[0]
            row_mask = (row[:-1] == mins)
            if exclude_last:
                row_mask = row_mask & (hyp_idx < hyp_lens)
            else:
                row_mask = row_mask & (hyp_idx <= hyp_lens)
            mask[hyp_idx] = row_mask
        elif return_prf_dsts and (hyp_idx < max_hyp_steps or not exclude_last):
            prefix_ers[hyp_idx] = row[ref_lens, batch_range]
    if return_mask:
        mask = mask & (
            (torch.arange(max_ref_steps, device=device)
                .unsqueeze(1)
                .expand(max_ref_steps, batch_size) < ref_lens)
            .unsqueeze(0)
        )
        return mask
    elif return_prf_dsts:
        if norm:
            prefix_ers = prefix_ers / ref_lens.float()
            zero_mask = ref_lens.eq(0).unsqueeze(0)
            if zero_mask.any():
                if warn:
                    warnings.warn(
                        "ref contains empty transcripts. Error rates will be "
                        "0 for prefixes of length 0, 1 otherwise. To suppress "
                        "this warning, set warn=False"
                    )
                prefix_ers = torch.where(
                    zero_mask,
                    (
                        torch.arange(prefix_ers.shape[0], device=device)
                        .gt(0).float().unsqueeze(1).expand_as(prefix_ers)
                    ),
                    prefix_ers,
                )
        prefix_ers = prefix_ers.masked_fill(
            (
                torch.arange(prefix_ers.shape[0], device=device).unsqueeze(1)
                .ge(hyp_lens + (0 if exclude_last else 1))
            ),
            padding,
        )
        if batch_first:
            prefix_ers = prefix_ers.t()
        return prefix_ers
    er = row[ref_lens, batch_range]
    if norm:
        er = er / ref_lens.float()
        zero_mask = ref_lens.eq(0)
        if zero_mask.any():
            if warn:
                warnings.warn(
                    "ref contains empty transcripts. Error rates for entries "
                    "will be 1 if any insertion and 0 otherwise. To suppress "
                    "this warning, set warn=False")
            er = torch.where(
                zero_mask,
                hyp_lens.gt(0).float(),
                er,
            )
    return er



# Copyright 2019 Sean Robertson

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import sys
import argparse
import math
import warnings

from collections import defaultdict, OrderedDict

import torch
import pydrobert.torch.data as data
import pydrobert.torch.util as util

__author__ = "Sean Robertson"
__email__ = "sdrobert@cs.toronto.edu"
__license__ = "Apache 2.0"
__copyright__ = "Copyright 2019 Sean Robertson"
__all__ = [
    'compute_torch_token_data_dir_error_rates',
    'ctm_to_torch_token_data_dir',
    'get_torch_spect_data_dir_info',
    'torch_token_data_dir_to_trn',
    'trn_to_torch_token_data_dir',
]


def _get_torch_spect_data_dir_info_parse_args(args):
    parser = argparse.ArgumentParser(
        description=get_torch_spect_data_dir_info.__doc__,
    )
    parser.add_argument('dir', type=str, help='The torch data directory')
    parser.add_argument(
        'out_file', nargs='?', type=argparse.FileType('w'),
        default=sys.stdout,
        help='The file to write to. If unspecified, stdout',
    )
    parser.add_argument(
        '--strict', action='store_true', default=False,
        help='If set, validate the data directory before collecting info. The '
        'process is described in pydrobert.torch.data.validate_spect_data_set'
    )
    parser.add_argument(
        '--file-prefix', default='',
        help='The file prefix indicating a torch data file'
    )
    parser.add_argument(
        '--file-suffix', default='.pt',
        help='The file suffix indicating a torch data file'
    )
    parser.add_argument(
        '--feat-subdir', default='feat',
        help='Subdirectory where features are stored'
    )
    parser.add_argument(
        '--ali-subdir', default='ali',
        help='Subdirectory where alignments are stored'
    )
    parser.add_argument(
        '--ref-subdir', default='ref',
        help='Subdirectory where reference token sequences are stored'
    )
    return parser.parse_args(args)


def get_torch_spect_data_dir_info(args=None):
    '''Write info about the specified SpectDataSet data dir

    A torch :class:`pydrobert.torch.data.SpectDataSet` data dir is of the
    form::

        dir/
            feat/
                <file_prefix><utt1><file_suffix>
                <file_prefix><utt2><file_suffix>
                ...
            [ali/
                <file_prefix><utt1><file_suffix>
                <file_prefix><utt1><file_suffix>
                ...
            ]
            [ref/
                <file_prefix><utt1><file_suffix>
                <file_prefix><utt1><file_suffix>
                ...
            ]

    Where ``feat`` contains :class:`torch.FloatTensor` of shape ``(N, F)``,
    where ``N`` is the number of frames (variable) and ``F`` is the number of
    filters (fixed), ``ali``, if there, contains :class:`torch.LongTensor` of
    shape ``(N,)`` indicating the appropriate class labels (likely pdf-ids for
    discriminative training in an DNN-HMM), and ``ref``, if there, contains
    :class:`torch.LongTensor` of shape ``(R, 3)`` indicating a sequence of
    reference tokens where element indexed by ``[i, 0]`` is a token id, ``[i,
    1]`` is the inclusive start frame of the token (or a negative value if
    unknown), and ``[i, 2]`` is the exclusive end frame of the token.

    This command writes the following space-delimited key-value pairs to an
    output file in sorted order:

    1. "max_ali_class", the maximum inclusive class id found over ``ali/``
       (if available, ``-1`` if not)
    2. "max_ref_class", the maximum inclussive class id found over ``ref/``
       (if available, ``-1`` if not)
    3. "num_utterances", the total number of listed utterances
    4. "num_filts", ``F``
    5. "total_frames", ``sum(N)`` over the data dir
    6. "count_<i>", the number of instances of the class "<i>" that appear
       in ``ali`` (if available). If "count_<i>" is a valid key, then so
       are "count_<0 to i>". "count_<i>" is left-padded with zeros to ensure
       that the keys remain in the same order in the table as the class
       indices.  The maximum ``i`` will be equal to ``maximum_ali_class``

    Note that the output can be parsed as a `Kaldi <http://kaldi-asr.org/>`__
    text table of integers.
    '''
    try:
        options = _get_torch_spect_data_dir_info_parse_args(args)
    except SystemExit as ex:
        return ex.code
    if not os.path.isdir(options.dir):
        print("'{}' is not a directory".format(options.dir), file=sys.stderr)
        return 1
    data_set = data.SpectDataSet(
        options.dir,
        file_prefix=options.file_prefix,
        file_suffix=options.file_suffix,
        feat_subdir=options.feat_subdir,
        ali_subdir=options.ali_subdir,
        ref_subdir=options.ref_subdir,
    )
    if options.strict:
        data.validate_spect_data_set(data_set)
    info_dict = {
        'num_utterances': len(data_set),
        'total_frames': 0,
        'max_ali_class': -1,
        'max_ref_class': -1,
    }
    counts = dict()
    for feat, ali, ref in data_set:
        info_dict['num_filts'] = feat.size()[1]
        info_dict['total_frames'] += feat.size()[0]
        if ali is not None:
            for class_idx in ali:
                class_idx = class_idx.item()
                if class_idx < 0:
                    raise ValueError('Got a negative ali class idx')
                info_dict['max_ali_class'] = max(
                    class_idx, info_dict['max_ali_class'])
                counts[class_idx] = counts.get(class_idx, 0) + 1
        if ref is not None:
            if ref.min().item() < 0:
                raise ValueError('Got a negative ref class idx')
            info_dict['max_ref_class'] = max(
                info_dict['max_ref_class'], ref.max().item())
    if info_dict['max_ali_class'] == 0:
        info_dict['count_0'] = counts[0]
    elif info_dict['max_ali_class'] > 0:
        count_fmt_str = 'count_{{:0{}d}}'.format(
            int(math.log10(info_dict['max_ali_class'])) + 1)
        for class_idx in range(info_dict['max_ali_class'] + 1):
            info_dict[count_fmt_str.format(class_idx)] = counts.get(
                class_idx, 0)
    info_list = sorted(info_dict.items())
    for key, value in info_list:
        options.out_file.write("{} {}\n".format(key, value))
    if options.out_file != sys.stdout:
        options.out_file.close()
    return 0


def _trn_to_torch_token_data_dir_parse_args(args):
    parser = argparse.ArgumentParser(
        description=trn_to_torch_token_data_dir.__doc__,
    )
    parser.add_argument(
        'trn', type=argparse.FileType('r'),
        help='The input trn file'
    )
    parser.add_argument(
        'token2id', type=argparse.FileType('r'),
        help='A file containing mappings from tokens (e.g. words or phones) '
        'to unique IDs. Each line has the format ``<token> <id>``. The flag '
        '``--swap`` can be used to swap the expected ordering (i.e. '
        '``<id> <token>``)'
    )
    parser.add_argument(
        'dir',
        help='The directory to store token sequences to. If the directory '
        'does not exist, it will be created'
    )
    parser.add_argument(
        '--alt-handler', default='error', choices=('error', 'first'),
        help='How to handle transcription alternates. If "error", error if '
        'the "trn" file contains alternates. If "first", always treat the '
        'alternate as canon'
    )
    parser.add_argument(
        '--file-prefix', default='',
        help='The file prefix indicating a torch data file'
    )
    parser.add_argument(
        '--file-suffix', default='.pt',
        help='The file suffix indicating a torch data file'
    )
    parser.add_argument(
        '--swap', action='store_true', default=False,
        help='If set, swaps the order of key and value in `token2id`'
    )
    return parser.parse_args(args)


def _parse_token2id(file, swap, return_swap):
    ret = dict()
    ret_swapped = dict()
    for line_no, line in enumerate(file):
        line = line.strip()
        if not line:
            continue
        ls = line.split()
        if len(ls) != 2 or not ls[1 - int(swap)].lstrip('-').isdigit():
            raise ValueError(
                'Cannot parse line {} of {}'.format(line_no + 1, file.name))
        key, value = ls
        key, value = (int(key), value) if swap else (key, int(value))
        if key in ret:
            warnings.warn(
                '{} line {}: "{}" already exists. Mapping will be ambiguous'
                ''.format(file.name, line_no + 1, key))
        if value in ret_swapped:
            warnings.warn(
                '{} line {}: "{}" already exists. Mapping will be ambiguous'
                ''.format(file.name, line_no + 1, value))
        ret[key] = value
        ret_swapped[value] = key
    return ret_swapped if return_swap else ret


def _save_transcripts_to_dir(
        transcripts, token2id, file_prefix, file_suffix, dir_,
        frame_shift_ms=None):
    if not os.path.isdir(dir_):
        os.makedirs(dir_)
    for utt_id, transcript in transcripts:
        tok = data.transcript_to_token(transcript, token2id, frame_shift_ms)
        path = os.path.join(dir_, file_prefix + utt_id + file_suffix)
        torch.save(tok, path)


def trn_to_torch_token_data_dir(args=None):
    '''Convert a NIST "trn" file to the specified SpectDataSet data dir

    A "trn" file is the standard transcription file without alignment
    information used in the `sclite
    <http://www1.icsi.berkeley.edu/Speech/docs/sctk-1.2/sclite.htm>`__ toolkit.
    It has the format::

        here is a transcription (utterance_a)
        here is another (utterance_b)

    This command reads in a "trn" file and writes its contents as token
    sequences compatible with the ``ref/`` directory of a
    :class:`pydrobert.torch.data.SpectDataSet`. See the command
    :func:`get_torch_spect_data_dir_info` (command line
    "get-torch-spect-data-dir-info") for more information on a
    :class:`pydrobert.torch.data.SpectDataSet`
    '''
    try:
        options = _trn_to_torch_token_data_dir_parse_args(args)
    except SystemExit as ex:
        return ex.code
    token2id = _parse_token2id(
        options.token2id, options.swap, options.swap)
    transcripts = data.read_trn(options.trn)
    # we manually search for alternates in a first pass, as we don't know what
    # filters users have on warnings
    for utt_id, transcript in transcripts:
        old_transcript = transcript[:]
        transcript[:] = []
        while len(old_transcript):
            x = old_transcript.pop(0)
            if len(x) == 3 and x[1] == -1:
                x = x[0]
            if isinstance(x, str):
                transcript.append(x)
            elif options.alt_handler == 'error':
                print(
                    'Cannot handle alternate in "{}"'.format(utt_id),
                    file=sys.stderr)
                return 1
            else:  # first
                x[0].extend(old_transcript)
                old_transcript = x[0]
    _save_transcripts_to_dir(
        transcripts, token2id, options.file_prefix, options.file_suffix,
        options.dir)
    return 0


def _torch_token_data_dir_to_trn_parse_args(args=None):
    parser = argparse.ArgumentParser(
        description=torch_token_data_dir_to_trn.__doc__)
    parser.add_argument(
        'dir', help='The directory to read token sequences from')
    parser.add_argument(
        'id2token', type=argparse.FileType('r'),
        help='A file containing the mappings from unique IDs to tokens (e.g. '
        'words or phones). Each line has the format ``<id> <token>``. The '
        'flag ``--swap`` can be used to swap the expected ordering (i.e. '
        '``<token> <id>``)'
    )
    parser.add_argument(
        "trn", type=argparse.FileType('w'),
        help='The "trn" file to write transcriptions to'
    )
    parser.add_argument(
        '--file-prefix', default='',
        help='The file prefix indicating a torch data file'
    )
    parser.add_argument(
        '--file-suffix', default='.pt',
        help='The file suffix indicating a torch data file'
    )
    parser.add_argument(
        '--swap', action='store_true', default=False,
        help='If set, swaps the order of key and value in `id2token`'
    )
    return parser.parse_args(args)


def _load_transcripts_from_data_dir(
        dir_, id2token, file_prefix, file_suffix, frame_shift_ms=None,
        strip_timing=False):
    fpl = len(file_prefix)
    neg_fsl = -len(file_suffix)
    utt_ids = sorted(
        x[fpl:neg_fsl]
        for x in os.listdir(dir_)
        if x.startswith(file_prefix) and
        x.endswith(file_suffix)
    )
    transcripts = []
    for utt_id in utt_ids:
        tok = torch.load(os.path.join(
            dir_, file_prefix + utt_id + file_suffix))
        transcript = data.token_to_transcript(tok, id2token, frame_shift_ms)
        for idx in range(len(transcript)):
            token = transcript[idx]
            if isinstance(token, tuple):
                token = token[0]
                if strip_timing:
                    transcript[idx] = token
            if isinstance(token, int) and id2token is not None:
                assert token not in id2token
                raise ValueError(
                    'Utterance "{}": ID "{}" could not be found in id2token'
                    ''.format(utt_id, token))
        transcripts.append((utt_id, transcript))
    return transcripts


def torch_token_data_dir_to_trn(args=None):
    '''Convert a SpectDataSet token data dir to a NIST trn file

    A "trn" file is the standard transcription file without alignment
    information used in the `sclite
    <http://www1.icsi.berkeley.edu/Speech/docs/sctk-1.2/sclite.htm>`_
    toolkit. It has the format::

        here is a transcription (utterance_a)
        here is another (utterance_b)

    This command scans the contents of a directory like ``ref/`` in a
    :class:`pydrobert.torch.data.SpectDataSet` and converts each such file into
    a transcription. Each such transcription is then written to a "trn" file.
    See the command :func:`get_torch_spect_data_dir_info` (command line
    "get-torch-spect-data-dir-info") for more information on a
    :class:`pydrobert.torch.data.SpectDataSet`
    '''
    try:
        options = _torch_token_data_dir_to_trn_parse_args(args)
    except SystemExit as ex:
        return ex.code
    if not os.path.isdir(options.dir):
        print('"{}" is not a directory'.format(options.dir), file=sys.stderr)
        return 1
    id2token = _parse_token2id(
        options.id2token, not options.swap, options.swap)
    transcripts = _load_transcripts_from_data_dir(
        options.dir, id2token, options.file_prefix, options.file_suffix)
    data.write_trn(transcripts, options.trn)
    return 0


def _ctm_to_torch_token_data_dir_parse_args(args):
    parser = argparse.ArgumentParser(
        description=ctm_to_torch_token_data_dir.__doc__)
    parser.add_argument(
        'ctm', type=argparse.FileType('r'),
        help='The "ctm" file to read token segments from')
    parser.add_argument(
        'token2id', type=argparse.FileType('r'),
        help='A file containing mappings from tokens (e.g. words or phones) '
        'to unique IDs. Each line has the format ``<token> <id>``. The flag '
        '``--swap`` can be used to swap the expected ordering (i.e. '
        '``<id> <token>``)'
    )
    parser.add_argument(
        'dir', help='The directory to store token sequences to. If the '
        'directory does not exist, it will be created'
    )
    parser.add_argument(
        '--file-prefix', default='',
        help='The file prefix indicating a torch data file'
    )
    parser.add_argument(
        '--file-suffix', default='.pt',
        help='The file suffix indicating a torch data file'
    )
    parser.add_argument(
        '--swap', action='store_true', default=False,
        help='If set, swaps the order of key and value in `token2id`'
    )
    parser.add_argument(
        '--frame-shift-ms', type=int, default=10,
        help='The number of milliseconds that have passed between consecutive '
        'frames. Used to convert between time in seconds and frame index'
    )
    utt_group = parser.add_mutually_exclusive_group()
    utt_group.add_argument(
        '--wc2utt', type=argparse.FileType('r'), default=None,
        help='A file mapping wavefile name and channel combinations (e.g. '
        '``utt_1 A``) to utterance IDs. Each line of the file has the format '
        '``<wavefile_name> <channel> <utt_id>``. If neither "--wc2utt" nor '
        '"--utt2wc" has been specied, the wavefile name will be treated as '
        'the utterance ID')
    utt_group.add_argument(
        '--utt2wc', type=argparse.FileType('r'), default=None,
        help='A file mapping utterance IDs to wavefile name and channel '
        'combinations (e.g. ``utt_1 A``). Each line of the file has the '
        'format ``<utt_id> <wavefile_name> <channel>``. If neither "--wc2utt" '
        'nor "--utt2wc" has been specied, the wavefile name will be treated '
        'as the utterance ID')
    return parser.parse_args(args)


def _parse_wc2utt(file, swap, return_swap):
    ret = dict()
    ret_swapped = dict()
    for line_no, line in enumerate(file):
        line = line.strip()
        if not line:
            continue
        ls = line.split()
        if len(ls) != 3:
            raise ValueError(
                'Cannot parse line {} of {}'.format(line_no + 1, file.name),
            )
        first, mid, last = ls
        key, value = ((mid, last), first) if swap else ((first, mid), last)
        if key in ret:
            warnings.warn(
                '{} line {}: "{}" already exists. Mapping will be '
                ''.format(file.name, line_no + 1, key)
            )
        if value in ret_swapped:
            warnings.warn(
                '{} line {}: "{}" already exists. Mapping will be '
                ''.format(file.name, line_no + 1, value)
            )
        ret[key] = value
        ret_swapped[value] = key
    return ret_swapped if return_swap else ret


def ctm_to_torch_token_data_dir(args=None):
    '''Convert a NIST "ctm" file to a SpectDataSet token data dir

    A "ctm" file is a transcription file with token alignments (a.k.a. a
    time-marked conversation file) used in the `sclite
    <http://www1.icsi.berkeley.edu/Speech/docs/sctk-1.2/sclite.htm>`_ toolkit.
    Here is the format::

        utt_1 A 0.2 0.1 hi
        utt_1 A 0.3 1.0 there  ;; comment
        utt_2 A 0.0 1.0 next
        utt_3 A 0.1 0.4 utterance

    Where the first number specifies the token start time (in seconds) and the
    second the duration.

    This command reads in a "ctm" file and writes its contents as token
    sequences compatible with the ``ref/`` directory of a
    :class:`pydrobert.torch.data.SpectDataSet`. See the command
    :func:`get_torch_spect_data_dir_info` (command line
    "get-torch-spect-data-dir-info") for more information on a
    :class:`pydrobert.torch.data.SpectDataSet`
    '''
    try:
        options = _ctm_to_torch_token_data_dir_parse_args(args)
    except SystemExit as ex:
        return ex.code
    token2id = _parse_token2id(
        options.token2id, options.swap, options.swap)
    if options.wc2utt:
        wc2utt = _parse_wc2utt(options.wc2utt, False, False)
    elif options.utt2wc:
        wc2utt = _parse_wc2utt(options.utt2wc, True, False)
    else:
        wc2utt = None
    transcripts = data.read_ctm(options.ctm, wc2utt)
    _save_transcripts_to_dir(
        transcripts, token2id, options.file_prefix, options.file_suffix,
        options.dir, options.frame_shift_ms)
    return 0


def _torch_token_data_dir_to_ctm_parse_args(args):
    parser = argparse.ArgumentParser(
        description=torch_token_data_dir_to_ctm.__doc__)
    parser.add_argument(
        'dir', help='The directory to read token sequences from')
    parser.add_argument(
        'id2token', type=argparse.FileType('r'),
        help='A file containing mappings from unique IDs to tokens (e.g. '
        'words or phones). Each line has the format ``<id> <token>``. The '
        '``--swap`` can be used to swap the expected ordering (i.e. '
        '``<token> <id>``)'
    )
    parser.add_argument(
        'ctm', type=argparse.FileType('w'),
        help='The "ctm" file to write token segments to')
    parser.add_argument(
        '--file-prefix', default='',
        help='The file prefix indicating a torch data file'
    )
    parser.add_argument(
        '--file-suffix', default='.pt',
        help='The file suffix indicating a torch data file'
    )
    parser.add_argument(
        '--swap', action='store_true', default=False,
        help='If set, swaps the order of key and value in `id2token`'
    )
    parser.add_argument(
        '--frame-shift-ms', type=int, default=10,
        help='The number of milliseconds that have passed between consecutive '
        'frames. Used to convert between time in seconds and frame index'
    )
    utt_group = parser.add_mutually_exclusive_group()
    utt_group.add_argument(
        '--wc2utt', type=argparse.FileType('r'), default=None,
        help='A file mapping wavefile name and channel combinations (e.g. '
        '``utt_1 A``) to utterance IDs. Each line of the file has the format '
        '``<wavefile_name> <channel> <utt_id>``')
    utt_group.add_argument(
        '--utt2wc', type=argparse.FileType('r'), default=None,
        help='A file mapping utterance IDs to wavefile name and channel '
        'combinations (e.g. ``utt_1 A``). Each line of the file has the '
        'format ``<utt_id> <wavefile_name> <channel>``')
    utt_group.add_argument(
        '--channel', default='A',
        help='If neither "--wc2utt" nor "--utt2wc" is specified, utterance '
        'IDs are treated as wavefile names and are given the value of this '
        'flag as a channel'
    )
    return parser.parse_args(args)


def torch_token_data_dir_to_ctm(args=None):
    '''Convert a SpectDataSet token data directory to a NIST "ctm" file

    A "ctm" file is a transcription file with token alignments (a.k.a. a
    time-marked conversation file) used in the `sclite
    <http://www1.icsi.berkeley.edu/Speech/docs/sctk-1.2/sclite.htm>`__ toolkit.
    Here is the format::

        utt_1 A 0.2 0.1 hi
        utt_1 A 0.3 1.0 there  ;; comment
        utt_2 A 0.0 1.0 next
        utt_3 A 0.1 0.4 utterance

    Where the first number specifies the token start time (in seconds) and the
    second the duration.

    This command scans the contents of a directory like ``ref/`` in a
    :class:`pydrobert.torch.data.SpectDataSet` and converts each such file into
    a transcription. Every token in a given transcription must have information
    about its duration. Each such transcription is then written to the "ctm"
    file. See the command :func:`get_torch_spect_data_dir_info` (command line
    "get-torch-spect-data-dir-info") for more information on a
    :class:`pydrobert.torch.data.SpectDataSet`
    '''
    try:
        options = _torch_token_data_dir_to_ctm_parse_args(args)
    except SystemExit as ex:
        return ex.code
    id2token = _parse_token2id(
        options.id2token, not options.swap, options.swap)
    if options.wc2utt:
        utt2wc = _parse_wc2utt(options.wc2utt, False, True)
    elif options.utt2wc:
        utt2wc = _parse_wc2utt(options.utt2wc, True, True)
    else:
        utt2wc = options.channel
    transcripts = _load_transcripts_from_data_dir(
        options.dir, id2token, options.file_prefix, options.file_suffix,
        options.frame_shift_ms)
    data.write_ctm(transcripts, options.ctm, utt2wc)
    return 0


def _compute_torch_token_data_dir_parse_args(args):
    parser = argparse.ArgumentParser(
        description=compute_torch_token_data_dir_error_rates.__doc__)
    parser.add_argument(
        'dir', help='If the `hyp` argument is not specified, this is the '
        'parent directory of two subdirectories, ``ref/`` and ``hyp/``, which '
        'contain the reference and hypothesis transcripts, respectively. If '
        'the ``--hyp`` argument is specified, this is the reference '
        'transcript directory'
    )
    parser.add_argument(
        'hyp', nargs='?', default=None,
        help='The hypothesis transcript directory',
    )
    parser.add_argument(
        'out', nargs='?', type=argparse.FileType('w'), default=sys.stdout,
        help='Where to print the error rate to. Defaults to stdout',
    )
    parser.add_argument(
        '--id2token', type=argparse.FileType('r'), default=None,
        help='A file containing mappings from unique IDs to tokens (e.g. '
        'words or phones). Each line has the format ``<id> <token>``. The '
        '``--swap`` flag can be used to swap the expected ordering (i.e. '
        '``<token> <id>``). ``--id2token`` can be used to collapse unique IDs '
        'together. Also, ``--ignore`` will contain a list of strings instead '
        'of IDs'
    )
    parser.add_argument(
        '--ignore', type=argparse.FileType('r'), default=None,
        help='A file containing a whitespace-delimited list of elements to '
        'ignore in both the reference and hypothesis transcripts. If '
        '``--id2token`` is specified, the file should contain words. If '
        '``--id2token`` is not specified, the file should contain IDs '
        '(integers)'
    )
    parser.add_argument(
        '--file-prefix', default='',
        help='The file prefix indicating a torch data file'
    )
    parser.add_argument(
        '--file-suffix', default='.pt',
        help='The file suffix indicating a torch data file'
    )
    parser.add_argument(
        '--swap', action='store_true', default=False,
        help='If set, swaps the order of key and value in `id2token`'
    )
    parser.add_argument(
        '--warn-missing', action='store_true', default=False,
        help='If set, warn and exclude any utterances that are missing either '
        'a reference or hypothesis transcript. The default is to error'
    )
    parser.add_argument(
        '--distances', action='store_true', default=False,
        help='If set, do not normalize by reference lengths'
    )
    parser.add_argument(
        '--per-utt', action='store_true', default=False,
        help='If set, return lines of ``<utt_id> <error_rate>`` denoting the '
        'per-utterance error rates instead of the average'
    )
    parser.add_argument(
        '--ins-cost', type=float, default=1.,
        help='The cost of an adding a superfluous token to a hypothesis '
        'transcript',
    )
    parser.add_argument(
        '--del-cost', type=float, default=1.,
        help='The cost of missing a token from a reference transcript',
    )
    parser.add_argument(
        '--sub-cost', type=float, default=1.,
        help='The cost of swapping a reference token with a hypothesis token',
    )
    parser.add_argument(
        '--batch-size', type=int, default=100,
        help='The number of error rates to compute at once. Reduce if you '
        'run into memory errors'
    )
    return parser.parse_args(args)


def compute_torch_token_data_dir_error_rates(args=None):
    '''Compute error rates between reference and hypothesis token data dirs

    This is a very simple script that computes and prints the error rates
    between the ``ref/`` (reference/gold standard) token sequences and ``hyp/``
    (hypothesis/generated) token sequences in a
    :class:`pydrobert.torch.data.SpectDataSet` directory. An error rate is
    merely a `Levenshtein Distance
    <https://en.wikipedia.org/wiki/Levenshtein_distance>`__ normalized to the
    reference sequence length.

    While convenient and accurate, this script has very few features. Consider
    pairing the command ``torch-token-data-dir-to-trn`` with `sclite
    <http://www1.icsi.berkeley.edu/Speech/docs/sctk-1.2/sclite.htm>`__ instead.

    Many tasks will ignore some tokens (e.g. silences) or collapse others (e.g.
    phones). Please consult a standard recipe (such as those in `Kaldi
    <http://kaldi-asr.org/>`__) before performing these computations
    '''
    try:
        options = _compute_torch_token_data_dir_parse_args(args)
    except SystemExit as ex:
        return ex.code
    if options.hyp:
        ref_dir, hyp_dir = options.dir, options.hyp
    else:
        ref_dir = os.path.join(options.dir, 'ref')
        hyp_dir = os.path.join(options.dir, 'hyp')
    if not os.path.isdir(ref_dir):
        print('"{}" is not a directory'.format(ref_dir), file=sys.stderr)
        return 1
    if not os.path.isdir(hyp_dir):
        print('"{}" is not a directory'.format(hyp_dir), file=sys.stderr)
        return 1
    if options.id2token:
        id2token = _parse_token2id(
            options.id2token, not options.swap, options.swap)
    else:
        id2token = None
    if options.ignore:
        ignore = set(options.ignore.read().strip().split())
        if id2token is None:
            try:
                ignore = {int(x) for x in ignore}
            except ValueError:
                raise ValueError(
                    'If --id2token is not set, all elements in "{}" must be '
                    'integers'.format(options.ignore.name))
    else:
        ignore = set()
    ref_transcripts = _load_transcripts_from_data_dir(
        ref_dir, id2token, options.file_prefix, options.file_suffix,
        strip_timing=True)
    hyp_transcripts = _load_transcripts_from_data_dir(
        hyp_dir, id2token, options.file_prefix, options.file_suffix,
        strip_timing=True)
    idx = 0
    while idx < max(len(ref_transcripts), len(hyp_transcripts)):
        missing_ref = missing_hyp = False
        if idx == len(ref_transcripts):
            missing_hyp = True
        elif idx == len(hyp_transcripts):
            missing_ref = True
        elif ref_transcripts[idx][0] < hyp_transcripts[idx][0]:
            missing_ref = True
        elif hyp_transcripts[idx][0] < ref_transcripts[idx][0]:
            missing_hyp = True
        if missing_hyp or missing_ref:
            if missing_hyp:
                fmt_tup = hyp_dir, hyp_transcripts[idx][0], ref_dir
                del hyp_transcripts[idx]
            else:
                fmt_tup = ref_dir, ref_transcripts[idx][0], hyp_dir
                del ref_transcripts[idx]
            msg = (
                'Directory "{}" contains utterance "{}" which directory "{}" '
                'does not contain'
            ).format(*fmt_tup)
            if options.warn_missing:
                warnings.warn(msg + '. Skipping')
            else:
                raise ValueError(msg)
        else:
            idx += 1
    assert len(ref_transcripts) == len(hyp_transcripts)
    idee_, eos, padding = [0], -1, -2

    def get_idee():
        v = idee_[0]
        idee_[0] += 1
        return v

    token2id = defaultdict(get_idee)
    error_rates = OrderedDict()
    while len(ref_transcripts):
        batch_ref_transcripts = ref_transcripts[:options.batch_size]
        batch_hyp_transcripts = hyp_transcripts[:options.batch_size]
        ref_transcripts = ref_transcripts[options.batch_size:]
        hyp_transcripts = hyp_transcripts[options.batch_size:]
        ref = torch.nn.utils.rnn.pad_sequence(
            [
                torch.tensor([
                    token2id[token] for token in transcript
                    if token not in ignore
                ] + [eos])
                for _, transcript in batch_ref_transcripts
            ],
            padding_value=padding,
        )
        hyp = torch.nn.utils.rnn.pad_sequence(
            [
                torch.tensor([
                    token2id[token] for token in transcript
                    if token not in ignore
                ] + [eos])
                for _, transcript in batch_hyp_transcripts
            ],
            padding_value=padding,
        )
        ers = util.error_rate(
            ref, hyp, eos=eos, include_eos=False, ins_cost=options.ins_cost,
            del_cost=options.del_cost, sub_cost=options.sub_cost,
            norm=not options.distances,
        )
        for (utt_id, _), er in zip(batch_ref_transcripts, ers):
            error_rates[utt_id] = er.item()
    if options.per_utt:
        for utt_id, er in error_rates.items():
            options.out.write('{} {}\n'.format(utt_id, er))
    else:
        options.out.write('{}\n'.format(
            sum(error_rates.values()) / len(error_rates)))
    return 0



# Copyright 2019 Sean Robertson

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''Common neural layers from the literature not included in pytorch.nn

Notes
-----
Though loss functions could be considered neural layers, because they are
specific to training, they are included in ``pydrobert.torch.training`` instead
'''

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import abc

import torch

from future.utils import with_metaclass

__author__ = "Sean Robertson"
__email__ = "sdrobert@cs.toronto.edu"
__license__ = "Apache 2.0"
__copyright__ = "Copyright 2019 Sean Robertson"
__all__ = [
    'ConcatSoftAttention',
    'DotProductSoftAttention',
    'GeneralizedDotProductSoftAttention',
    'GlobalSoftAttention',
    'MultiHeadedAttention',
]

# XXX(sdrobert): a quick note on style. pytorch doesn't tend to protect its
# read-only attributes using private members, so we do the same


class GlobalSoftAttention(with_metaclass(abc.ABCMeta, torch.nn.Module)):
    r'''Parent class for soft attention mechanisms on an entire input sequence

    Global soft attention mechansims [bahdanau2015]_ are a way of getting rid
    of one variable-length sequence dimension ``T`` in an input `key` using a
    weighted sum of a tensor `value` that is informed by some other tensor,
    `query`. The weights are dictated by the function ``score(query, key)``.
    Usually, this is in the context of encoder-decoder architectures, which is
    explained here.

    Assume `query` is a tensor of shape ``(num_batch, query_size)``
    representing a single hidden state of a decoder RNN. Assume `key` is a
    tensor of shape ``(T, num_batch, key_size)`` representing the encoder
    output, ``dim == 0`` to specify that the variable-length dimension of `key`
    is the zero-th dimension, and ``value == key``. The output `out` will be a
    tensor of shape ``(num_batch, key_size)``. Letting :math:`t` index the
    `dim`-th dimension:

        .. math::

            out = \sum_t a_t value_t

    ``a`` is the attention vector. In our example, ``a`` will be of shape
    ``(T, num_batch)``. ``a`` is the result of a softmax over the `dim`-th
    dimension of another tensor ``e`` of shape ``(T, num_batch)`` with an
    optional `mask`

    .. math::

        a = softmax(e * mask - (1 - mask) \infty, dim)

    `mask` (if specified) is of shape ``(T, num_batch)`` and will set ``a`` to
    zero wherever the mask is zero. `mask` can be used to indicate padded
    values when `key` consists of variable-length sequences.

    ``e`` is the result of a score function over `key` and `query`

    .. math::

        e = score(query, key)

    ``score()`` is implemented by subclasses of :class:`GlobalSoftAttention`

    The signature when calling an instance this module is:

        attention(query, key, value[, mask])

    Parameters
    ----------
    query_size : int
        The length of the last dimension of the `query` argument
    key_size : int
        The length of the last dimension of the `key` argument
    dim : int, optional
        The sequence dimension of the `key` argument

    Attributes
    ----------
    query_size, key_size, dim : int

    Examples
    --------

    A simple auto-regressive decoder using soft attention on encoder outputs
    with "concat"-style attention

    >>> T, num_batch, encoded_size, hidden_size = 100, 5, 30, 124
    >>> num_classes, start, eos, max_decoder_steps = 20, -1, 0, 100
    >>> encoded_lens = torch.randint(1, T + 1, (num_batch,))
    >>> len_mask = torch.where(
    ...     torch.arange(T).unsqueeze(-1) < encoded_lens,
    ...     torch.tensor(1),
    ...     torch.tensor(0),
    ... )
    >>> encoded = torch.randn(T, num_batch, encoded_size)
    >>> rnn = torch.nn.RNNCell(encoded_size + 1, hidden_size)
    >>> ff = torch.nn.Linear(hidden_size, num_classes)
    >>> attention = ConcatSoftAttention(hidden_size, encoded_size)
    >>> h = torch.zeros((num_batch, hidden_size))
    >>> y = torch.full((1, num_batch), -1, dtype=torch.long)
    >>> for _ in range(max_decoder_steps):
    >>>     if y[-1].eq(eos).all():
    >>>         break
    >>>     context = attention(h, encoded, encoded, len_mask)
    >>>     cat = torch.cat([context, y[-1].unsqueeze(-1).float()], 1)
    >>>     h = rnn(cat)
    >>>     logit = ff(h)
    >>>     y_next = logit.argmax(-1).masked_fill(y[-1].eq(eos), eos)
    >>>     y = torch.cat([y, y_next.unsqueeze(0)], 0)

    See Also
    --------
    :ref:`Advanced Attention and Transformer Networks`
        :class:`GlobalSoftAttention` is compatible with a variety of inputs.
        This tutorial gives a toy transformer network to illustrate
        broadcasting semantics
    '''

    def __init__(self, query_size, key_size, dim=0):
        super(GlobalSoftAttention, self).__init__()
        self.query_size = query_size
        self.key_size = key_size
        self.dim = dim

    @abc.abstractmethod
    def score(self, query, key):
        '''Calculate the score function over the entire input

        This is implemented by subclasses of :class:`GlobalSoftAttention`

        ``query.unsqueeze(self.dim)[..., 0]`` broadcasts with ``value[...,
        0]``. The final dimension of `query` is of length ``self.query_size``
        and the final dimension of `key` should be of length ``self.key_size``

        Parameters
        ----------
        query : torch.FloatTensor
        key : torch.FloatTensor

        Returns
        -------
        e : torch.FloatTensor
            Of the same shape as the above broadcasted tensor
        '''
        raise NotImplementedError()

    def forward(self, query, key, value, mask=None):
        if query.dim() + 1 != key.dim():
            raise ValueError('query must have one fewer dimension than key')
        if key.dim() != value.dim():
            raise ValueError(
                "key must have same number of dimensions as value")
        if query.shape[-1] != self.query_size:
            raise ValueError('Last dimension of query must match query_size')
        if key.shape[-1] != self.key_size:
            raise ValueError('Last dimension of key must match key_size')
        key_dim = key.dim()
        if self.dim > key_dim - 2 or self.dim < -key_dim + 1:
            raise ValueError(
                'dim must be in the range [{}, {}]'
                ''.format(-key_dim + 1, key_dim - 2)
            )
        e = self.score(query, key)
        if mask is not None:
            e = e.masked_fill(mask.eq(0), -float('inf'))
        a = torch.nn.functional.softmax(e, self.dim)
        c = (a.unsqueeze(-1) * value).sum(self.dim)
        return c

    def extra_repr(self):
        return 'query_size={}, key_size={}, dim={}'.format(
            self.query_size, self.key_size, self.dim)

    def reset_parameters(self):
        pass


class DotProductSoftAttention(GlobalSoftAttention):
    r'''Global soft attention with dot product score function

    From [luong2015]_, the score function for this attention mechanism is

    .. math::

        e = scale\_factor \sum_i query_i key_i

    Where :math:`i` indexes the last dimension of both the query and key

    Parameters
    ----------
    size : int
        Both the query and key size
    dim : int, optional
    scale_factor : float, optional
        A floating point to multiply the each :math:`e` with. Usually
        1, but if set to :math:`1 / size`, you'll get the scaled dot-product
        attention of [vaswani2017]_

    Attributes
    ----------
    query_size, key_size, dim : int
    scale_factor : float

    See Also
    --------
    GlobalSoftAttention
        For a description of how to call this module, how it works, etc.
    '''

    def __init__(self, size, dim=0, scale_factor=1.):
        super(DotProductSoftAttention, self).__init__(size, size, dim)
        self.scale_factor = scale_factor

    def score(self, query, key):
        query = query.unsqueeze(self.dim)
        return (query * key).sum(-1) * self.scale_factor

    def extra_repr(self):
        return 'size={}, dim={}'.format(self.query_size, self.dim)


class GeneralizedDotProductSoftAttention(GlobalSoftAttention):
    r'''Dot product soft attention with a learned matrix in between

    The "general" score function from [luong2015]_, the score function for this
    attention mechanism is

    .. math::

        e = \sum_q query_q \sum_k W_{qk} key_k

    For some learned matrix :math:`W`. :math:`q` indexes the last dimension
    of `query` and :math:`k` the last dimension of `key`

    Parameters
    ----------
    query_size : int
    key_size : int
    dim : int, optional
    bias : bool, optional
        Whether to add a bias term ``b``: :math:`W key + b`

    Attributes
    ----------
    query_size, key_size, dim : int
    W : torch.nn.Linear
        The matrix :math:`W`

    See Also
    --------
    GlobalSoftAttention
        For a description of how to call this module, how it works, etc.
    '''

    def __init__(self, query_size, key_size, dim=0, bias=False):
        super(GeneralizedDotProductSoftAttention, self).__init__(
            query_size, key_size, dim)
        self.W = torch.nn.Linear(key_size, query_size, bias=bias)

    def score(self, query, key):
        Wkey = self.W(key)
        query = query.unsqueeze(self.dim)
        return (query * Wkey).sum(-1)

    def reset_parameters(self):
        self.W.reset_parameters()


class ConcatSoftAttention(GlobalSoftAttention):
    r'''Attention where query and key are concatenated, then fed into an MLP

    Proposed in [luong2015]_, though quite similar to that proposed in
    [bahdanau2015]_, the score function for this layer is:

    .. math::

        e = \sum_i v_i \tanh(\sum_c W_{ic} [query, key]_c)

    For some learned matrix :math:`W` and vector :math:`v`, where
    :math:`[query, key]` indicates concatenation along the last axis. `query`
    and `key` will be expanded to fit their broadcast dimensions. :math:`W`
    has shape ``(inter_size, key_size)`` and :math:`v` has shape
    ``(hidden_size,)``

    Parameters
    ----------
    query_size : int
    key_size : int
    dim : int, optional
    bias : bool, optional
        Whether to add bias term ``b`` :math:`W [query, key] + b`
    hidden_size : int, optional

    Attributes
    ----------
    query_size, key_size, dim, hidden_size : int
    W : torch.nn.Linear
        The matrix :math:`W`
    v : torch.nn.Linear
        The vector :math:`v` as a single-row matrix

    See Also
    --------
    GlobalSoftAttention
        For a description of how to call this module, how it works, etc.
    '''

    def __init__(
            self, query_size, key_size, dim=0, bias=False, hidden_size=1000):
        super(ConcatSoftAttention, self).__init__(query_size, key_size, dim)
        self.hidden_size = hidden_size
        self.W = torch.nn.Linear(
            query_size + key_size, hidden_size, bias=bias)
        # there's no point in a bias for v. It'll just be absorbed by the
        # softmax later. You could add a bias after the tanh layer, though...
        self.v = torch.nn.Linear(hidden_size, 1, bias=False)

    def score(self, query, key):
        query = query.unsqueeze(self.dim)
        query_wo_last, key_wo_last = torch.broadcast_tensors(
            query[..., 0], key[..., 0])
        query, _ = torch.broadcast_tensors(query, query_wo_last.unsqueeze(-1))
        key, _ = torch.broadcast_tensors(key, key_wo_last.unsqueeze(-1))
        cat = torch.cat([query, key], -1)
        Wcat = self.W(cat)
        return self.v(Wcat).squeeze(-1)

    def reset_parameters(self):
        self.W.reset_parameters()
        self.v.reset_parameters()

    def extra_repr(self):
        s = super(ConcatSoftAttention, self).extra_repr()
        s += ', hidden_size={}'.format(self.hidden_size)
        return s


class MultiHeadedAttention(torch.nn.Module):
    r'''Perform attention over a number of heads, concatenate, and project

    Multi-headed attention was proposed in [vaswani2017]_. It can be considered
    a wrapper around standard :class:`GlobalSoftAttention` that results in a
    similar output as a regular attention layer. The idea is to replicate
    transformed versions of the `query`, `key`, and `value` `num_heads` times.
    Letting :math:`h` index the head:

    .. math::

        query_h = W^Q_h query \\
        key_h = W^K_h key \\
        value_h = W^V_h value

    If `query` is of shape ``(..., query_size)``, :math:`W^Q_h` is a learned
    matrix of shape ``(query_size, d_q)`` that acts on the final dimension of
    `query`. Likewise, :math:`W^K_h` is of shape ``(key_size, d_k)`` and
    :math:`W^V_h` is of shape ``(value_size, d_v)``.

    Each head is then determined via a wrapped :class:`GlobalSoftAttention`
    instance, `single_head_attention`:

    .. math::

        head_h = single\_head\_attention(query_h, key_h, value_h, mask)

    Where `mask` is repeated over all :math:`h`.

    Since each :math:`head_h` has the same shape, they can be concatenated
    along the last dimension to get the tensor :math:`cat` of shape
    ``(..., d_v * num_heads)``, which is linearly transformed into the output

    .. math::

        out = W^C cat

    With a learnable matrix :math:`W^C` of shape ``(d_v * num_heads,
    out_size)``. `out` has a shape ``(..., out_size)``

    This module has the following signature when called

        attention(query, key, value[, mask])

    Parameters
    ----------
    query_size : int
        The size of the last dimension of the `query` being passed to this
        module (not the size of a head's query)
    key_size : int
        The size of the last dimension of the `key` being passed to this
        module (not the size of a head's key)
    value_size : int
        The size of the last dimension of the `value` being passed to this
        module (not the size of a head's value)
    num_heads : int
        The number of heads to spawn
    single_head_attention : GlobalSoftAttention
        An instance of a subclass of :class:`GlobalSoftAttention` responsible
        for processing a head. `single_head_attention` attention will be used
        to derive the sequence dimension (``dim``) of `key` via
        ``single_head_attention.dim``, the size of a head's query ``d_k`` via
        ``single_head_attention.query_size``, and the size of a head's key via
        ``single_head_attention.key_size``
    out_size : int, optional
        The size of the last dimension of `out`. If unset, the default is to
        match `value_size`
    d_v : int, optional
        The size of the last dimension of a head's value. If unset, will
        default to ``max(1, value_size // num_heads)``
    bias_WQ : bool, optional
        Whether to add a bias term to :math:`W^Q`
    bias_WK : bool, optional
        Whether to add a bias term to :math:`W^K`
    bias_WV : bool, optional
        Whether to add a bias term to :math:`W^V`
    bias_WC : bool, optional
        Whether to add a bias term to :math:`W^C`

    Attributes
    ----------
    query_size, key_size, value_size, out_size, num_heads, dim : int
    d_q, d_k, d_v : int
    single_head_attention : GlobalSoftAttention
    WQ, WK, WV, WC : torch.nn.Linear
        Matrices :math:`W^Q`, :math:`W^K`, :math:`W^V`, and :math:`W^C`
    '''

    def __init__(
            self, query_size, key_size, value_size, num_heads,
            single_head_attention, out_size=None, d_v=None,
            bias_WQ=False, bias_WK=False, bias_WV=False, bias_WC=False):
        super(MultiHeadedAttention, self).__init__()
        self.query_size = query_size
        self.key_size = key_size
        self.value_size = value_size
        self.out_size = value_size if out_size is None else out_size
        self.num_heads = num_heads
        self.single_head_attention = single_head_attention
        # we don't keep these in sync in case someone's using
        # single_head_attention
        self.dim = single_head_attention.dim
        self.d_q = single_head_attention.query_size
        self.d_k = single_head_attention.key_size
        self.d_v = max(1, value_size // num_heads) if d_v is None else d_v
        self.WQ = torch.nn.Linear(
            query_size, num_heads * self.d_q, bias=bias_WQ)
        self.WK = torch.nn.Linear(
            key_size, num_heads * self.d_k, bias=bias_WK)
        self.WV = torch.nn.Linear(
            value_size, num_heads * self.d_v, bias=bias_WV)
        self.WC = torch.nn.Linear(
            self.d_v * num_heads, self.out_size, bias=bias_WC)
        single_head_attention.reset_parameters()

    def forward(self, query, key, value, mask=None):
        query_shape = tuple(query.shape)
        key_shape = tuple(key.shape)
        value_shape = tuple(value.shape)
        # we check dim here because we want to turn it into a positive
        # index before passing it to the head. This is because there's going
        # to be one more dimension to query, key, and value at the end that
        # represents the `num_heads` dimension, and we don't want negative
        # indices being wrongly offset because of it
        key_dim = len(key_shape)
        if self.dim > key_dim - 2 or self.dim < -key_dim + 1:
            raise ValuerError(
                'dim must be in the range [{}, {}]'
                ''.format(-key_dim + 1, key_dim - 2)
            )
        dim = (self.dim + key_dim) % key_dim
        query_heads = (
            self.WQ(query)
                .view(*(query_shape[:-1] + (self.num_heads, self.d_q)))
        )
        key_heads = (
            self.WK(key)
                .view(*(key_shape[:-1] + (self.num_heads, self.d_k)))
        )
        value_heads = (
            self.WV(value)
                .view(*(value_shape[:-1] + (self.num_heads, self.d_v)))
        )
        if mask is not None:
            mask = mask.unsqueeze(-2)
        old_dim = self.single_head_attention.dim
        try:
            self.single_head_attention.dim = dim
            cat = self.single_head_attention(
                query_heads, key_heads, value_heads, mask)
        finally:
            self.single_head_attention.dim = old_dim
        cat = cat.view(*(tuple(cat.shape[:-2]) + (self.num_heads * self.d_v,)))
        return self.WC(cat)

    def reset_parameters(self):
        self.WQ.reset_parameters()
        self.WK.reset_parameters()
        self.WV.reset_parameters()
        self.WC.reset_parameters()
        self.single_head_attention.reset_parameters()

    def extra_repr(self):
        return (
            'query_size={}, key_size={}, value_size={}, out_size={}, dim={}, '
            'num_heads={}, d_q={}, d_k={}, d_v={}'.format(
                self.query_size, self.key_size, self.value_size, self.out_size,
                self.dim, self.num_heads, self.d_q, self.d_k, self.d_v)
        )
